import py
import os
import sys
import pytest
from cffi import FFI
from array import array

if sys.platform != 'win32':

    ffi = FFI()
    ffi.cdef("""
    //void dump_all_known_symbols(int fd);
    int test_extract(char ** name, int * lineno, char ** src);
    int test_extract_sofile(char ** name, int * lineno, char ** src);
    """)
    with open("src/symboltable.c", "rb") as fd:
        source = fd.read().decode()
    source += """

    static char gname[64];
    static char gsrc[128];
    int test_extract(char ** name, int * lineno, char ** src)
    {
        *name = gname;
        *src = gsrc;
        gname[0] = 0;
        gsrc[0] = 0;
        return vmp_resolve_addr(&vmp_resolve_addr, gname, 64,
                                lineno, gsrc, 128);
    }
    int test_extract_sofile(char ** name, int * lineno, char ** src)
    {
        *name = gname;
        *src = gsrc;
        gname[0] = 0;
        gsrc[0] = 0;
        return vmp_resolve_addr(&abs, gname, 64,
                                lineno, gsrc, 128);
    }
    """
    libs = [] #['unwind', 'unwind-x86_64']
    includes = ['src']
    if sys.platform.startswith('linux'):
        for src in ["src/libbacktrace/state.c",
                    "src/libbacktrace/backtrace.c",
                    "src/libbacktrace/fileline.c",
                    "src/libbacktrace/posix.c",
                    "src/libbacktrace/mmap.c",
                    "src/libbacktrace/mmapio.c",
                    "src/libbacktrace/elf.c",
                    "src/libbacktrace/dwarf.c",
                    "src/libbacktrace/sort.c",
                   ]:
            with open(src, "rb") as fd:
                source += fd.read().decode()
        includes.append('src/libbacktrace')

    extra_compile = []
    if sys.platform.startswith("linux"):
        extra_compile = ['-DVMPROF_LINUX', '-DVMPROF_UNIX', '-w', '-g']

    # trick: compile with _CFFI_USE_EMBEDDING=1 which will not define Py_LIMITED_API
    ffi.set_source("vmprof.test._test_symboltable", source, include_dirs=includes,
                         define_macros=[('_CFFI_USE_EMBEDDING',1),('_PY_TEST',1)], libraries=libs,
                         extra_compile_args=extra_compile)

    ffi.compile(verbose=True)

@py.test.mark.skipif("sys.platform == 'win32'")
class TestSymbolTable(object):
    def setup_class(cls):
        from vmprof.test import _test_symboltable as clib
        cls.lib = clib.lib
        cls.ffi = clib.ffi

    def test_resolve_addr(self):
        lib = self.lib
        ffi = self.ffi
        name = ffi.new("char**")
        src = ffi.new("char**")
        _lineno = ffi.new("int*")
        lib.test_extract(name, _lineno, src)

        assert ffi.string(name[0]) == b"vmp_resolve_addr"
        assert ffi.string(src[0]).endswith(b"vmprof/test/_test_symboltable.c")
        # lines are not included in stab
        if sys.platform.startswith('linux'):
            with open("vmprof/test/_test_symboltable.c", "rb") as fd:
                lineno = 1
                for line in fd.readlines():
                    if "int vmp_resolve_addr(void * addr," in line.decode():
                        if _lineno[0] == lineno:
                            break
                    lineno += 1
                else:
                    assert False, "could not match line number"

    def test_sofile_in_srcfile(self):
        lib = self.lib
        ffi = self.ffi
        name = ffi.new("char**")
        src = ffi.new("char**")
        _lineno = ffi.new("int*")
        # the idea of this test is to extract some details out of e.g. libc
        # usually linux distros do not ship dwarf information unless you install
        # them.
        lib.test_extract_sofile(name, _lineno, src)

        assert ffi.string(name[0]) == b"abs"
        # should be something like /lib64/libc.so.6 (e.g. on Fedora 25)
        if sys.platform.startswith("linux"):
            assert b"libc" in ffi.string(src[0])
            assert b".so" in ffi.string(src[0])
        elif sys.platform == "darwin":
            # osx
            assert b"libsystem_c.dylib" in ffi.string(src[0])



import vmprof
import tempfile

from vmprof.stats import Stats
from vmprof.reader import _read_prof


class VMProfError(Exception):
    pass

class ProfilerContext(object):
    done = False

    def __init__(self, name, period, memory, native, real_time):
        if name is None:
            self.tmpfile = tempfile.NamedTemporaryFile("w+b", delete=False)
        else:
            self.tmpfile = open(name, "w+b")
        self.filename = self.tmpfile.name
        self.period = period
        self.memory = memory
        self.native = native
        self.real_time = real_time

    def __enter__(self):
        vmprof.enable(self.tmpfile.fileno(), self.period, self.memory,
                      native=self.native, real_time=self.real_time)

    def __exit__(self, type, value, traceback):
        vmprof.disable()
        self.tmpfile.close() # flushes the stream
        self.done = True


def read_profile(prof_file):
    file_to_close = None
    if not hasattr(prof_file, 'read'):
        prof_file = file_to_close = open(str(prof_file), 'rb')

    state = _read_prof(prof_file)

    if file_to_close:
        file_to_close.close()

    jit_frames = {}
    d = dict(state.virtual_ips)
    s = Stats(state.profiles, d, jit_frames, interp=state.interp_name,
              start_time=state.start_time, end_time=state.end_time,
              meta=state.meta, state=state)
    return s


class Profiler(object):
    ctx = None

    def __init__(self):
        self._lib_cache = {}

    def measure(self, name=None, period=0.001, memory=False, native=False, real_time=False):
        self.ctx = ProfilerContext(name, period, memory, native, real_time)
        return self.ctx

    def get_stats(self):
        if not self.ctx:
            raise VMProfError("no profiling done")
        if not self.ctx.done:
            raise VMProfError("profiling in process")
        res = read_profile(self.ctx.tmpfile.name)
        self.ctx = None
        return res




import tempfile

from vmprof import cli


def ini_config(content):
    _, name = tempfile.mkstemp()
    f = open(name, 'w')
    f.write(content)
    f.close()
    return name


def test_parser_config():
    test_file = ini_config("""
[global]
period = 10.0
    """)

    args = cli.parse_args([
        '--config', test_file,
        'example.py'
    ])

    assert args.period == 10.0


def test_parser_arg_precedence():
    test_file = ini_config("""
[global]
period = 10.0
web-url = example.com
    """)

    args = cli.parse_args([
        '--config', test_file,
        '--period', '5.0',
        'example.py'
    ])

    assert args.period == 5.0
    assert args.web is False
    assert args.web_url == "example.com"
    assert args.mem == False


def test_parser_without_section():

    test_file = ini_config("""
[global]
period = 0.1
web-url=example.com
enable-nonvirtual = False
no-native = True
    """)

    args = cli.parse_args([
        '--config', test_file,
        'example.py'
    ])

    assert test_file == args.config.name
    assert args.no_native == True



import sys
import pytest
import vmprof
from cffi import FFI
from array import array

sample = None
@pytest.mark.skipif("sys.platform == 'win32'")
class TestStack(object):
    def setup_class(cls):
        stack_ffi = FFI()
        stack_ffi.cdef("""
        typedef intptr_t ptr_t;
        int vmp_binary_search_ranges(ptr_t ip, ptr_t * l, int count);
        int vmp_ignore_ip(ptr_t ip);
        int vmp_ignore_symbol_count(void);
        ptr_t * vmp_ignore_symbols(void);
        void vmp_set_ignore_symbols(ptr_t * symbols, int count);
        int vmp_read_vmaps(const char * fname);
        void vmp_native_disable();
        """)
        with open("src/vmp_stack.c", "rb") as fd:
            source = fd.read().decode()
            libs = [] #['unwind', 'unwind-x86_64']
            if sys.platform.startswith('linux'):
                libs = ['unwind', 'unwind-x86_64']
            # trick: compile with _CFFI_USE_EMBEDDING=1 which will not define Py_LIMITED_API
            stack_ffi.set_source("vmprof.test._test_stack", source, include_dirs=['src'],
                                 define_macros=[('_CFFI_USE_EMBEDDING',1), ('PY_TEST',1),
                                                ('VMP_SUPPORTS_NATIVE_PROFILING',1)],
                                 libraries=libs, extra_compile_args=['-w', '-g'])

        stack_ffi.compile(verbose=True)
        from vmprof.test import _test_stack as clib
        cls.lib = clib.lib
        cls.ffi = clib.ffi

    def test_binary_search_ranges(self):
        a = array('l', [1,2,10,20,30,40])
        count = len(a)
        abuf = self.ffi.from_buffer(a)
        a_addr = self.ffi.cast("ptr_t*", abuf)
        assert self.lib.vmp_binary_search_ranges(0, a_addr, count) == -1
        assert self.lib.vmp_binary_search_ranges(1, a_addr, count) == 0
        assert self.lib.vmp_binary_search_ranges(2, a_addr, count) == 0
        for i in range(3,10):
            assert self.lib.vmp_binary_search_ranges(i, a_addr, count) == 0
        for i in range(10,21):
            assert self.lib.vmp_binary_search_ranges(i, a_addr, count) == 2
        for i in range(21,30):
            assert self.lib.vmp_binary_search_ranges(i, a_addr, count) == 2 
        for i in range(30,41):
            assert self.lib.vmp_binary_search_ranges(i, a_addr, count) == 4
        assert self.lib.vmp_binary_search_ranges(41, a_addr, count) == -1

    def test_ignore_ip(self):
        a = array('l', [1,2,100,150])
        abuf = self.ffi.from_buffer(a)
        a_addr = self.ffi.cast("ptr_t*", abuf)
        self.lib.vmp_set_ignore_symbols(a_addr, len(a))
        assert self.lib.vmp_ignore_ip(0) == 0
        assert self.lib.vmp_ignore_ip(1) == 1
        assert self.lib.vmp_ignore_ip(2) == 1
        assert self.lib.vmp_ignore_ip(3) == 0
        for i in range(3,100):
            assert self.lib.vmp_ignore_ip(i) == 0
        for i in range(100,151):
            assert self.lib.vmp_ignore_ip(i) == 1
        assert self.lib.vmp_ignore_ip(151) == 0
        self.lib.vmp_set_ignore_symbols(self.ffi.NULL, 0)

    @pytest.mark.skipif("not sys.platform.startswith('linux')")
    def test_read_vmaps(self, tmpdir):
        lib = self.lib
        f1 = tmpdir.join("vmap1")
        f1.write("""\
00000-11111 x y z w python
11111-22222 x y z w python site-packages
""")
        filename = str(f1).encode('utf-8')
        assert lib.vmp_read_vmaps(filename) == 1

        assert lib.vmp_ignore_symbol_count() == 2
        symbols = lib.vmp_ignore_symbols()
        assert symbols[0] == 0 and symbols[1] == 0x22222
        assert lib.vmp_ignore_ip(0x1) == 1
        assert lib.vmp_ignore_ip(0x11111) == 1
        assert lib.vmp_ignore_ip(0x11112) == 1
        assert lib.vmp_ignore_ip(0x22222) == 1
        assert lib.vmp_ignore_ip(0x22223) == 0
        lib.vmp_native_disable()

    @pytest.mark.skipif("not sys.platform.startswith('linux')")
    def test_self_vmaps(self):
        lib = self.lib
        assert lib.vmp_read_vmaps(b"/proc/self/maps") == 1

        count = lib.vmp_ignore_symbol_count()
        assert count >= 2
        symbols = lib.vmp_ignore_symbols()
        min = -1
        for i in range(0, count, 2):
            start, end = symbols[i], symbols[i+1]
            assert min < start
            assert start <= end
            min = end
        lib.vmp_native_disable()


    @pytest.mark.skipif("not sys.platform.startswith('darwin')")
    def test_read_vmaps_darwin(self):
        assert self.lib.vmp_read_vmaps(self.ffi.NULL) == 1

    @pytest.mark.skipif("not sys.platform.startswith('linux')")
    def test_overflow_vmaps(self, tmpdir):
        lib = self.lib
        f1 = tmpdir.join("vmap1")
        lines = []
        for l in range(0,10000, 2):
            # non overlapping ranges that must be considered! 
            lines.append('%x-%x x y z w python' % (l,l+1))
        f1.write('\n'.join(lines))
        filename = str(f1).encode('utf-8')
        assert lib.vmp_read_vmaps(filename) == 1
        for l in range(0, 10000):
            assert self.lib.vmp_ignore_ip(l) == 1
        assert self.lib.vmp_ignore_ip(10001) == 0



import six
from vmprof.reader import AssemblerCode, JittedCode, NativeCode

class EmptyProfileFile(Exception):
    pass

class Stats(object):
    def __init__(self, profiles, adr_dict=None, jit_frames=None, interp=None,
                 meta=None, start_time=None, end_time=None, state=None):
        self.profiles = profiles
        self.adr_dict = adr_dict
        self.functions = {}
        # kludgy, state is optional. stats should only take state as input
        if state:
            self.profile_lines = state.profile_lines
            self.profile_memory = state.profile_memory
        else:
            # unkown, for tests only
            self.profile_lines = False
            self.profile_memory = False
        self.generate_top()
        if jit_frames is None:
            jit_frames = set()
        if interp is None:
            interp = 'pypy' # why not
        self.interp = interp
        self.jit_frames = jit_frames
        self.meta = meta or {}
        self.start_time = start_time
        self.end_time = end_time

    def get_runtime_in_microseconds(self):
        if self.start_time is None or self.end_time is None:
            return 0 # old versions that do not emit start or end time
        ts = self.end_time - self.start_time
        return ts.total_seconds() * 1000000

    def get_name(self, addr):
        if addr not in self.adr_dict:
            return "unknown"
        line = self.adr_dict[addr]
        return line.split(":")[1]

    def find_addrs_containing_name(self, part):
        for adr, name in self.adr_dict.items():
            n, symbol, _, _ = name.split(':', 3)
            if part in symbol:
                yield adr

    def get_addr_info(self, addr):
        name = self.adr_dict.get(addr, None)
        if not name:
            return None
        lang, symbol, line, file = name.split(':', 3)
        return lang, symbol, line, file

    def getargv(self):
        return self.meta.get('argv', '')

    def getmeta(self, key, default):
        return self.meta.get(key, default) 

    def display(self, no):
        prof = self.profiles[no][0]
        return [self._get_name(elem) for elem in prof]

    def generate_top(self):
        for profile in self.profiles:
            current_iter = {}
            for i, addr in enumerate(profile[0]):
                if self.profile_lines and i % 2 == 1:
                    # this entry in the profile is a negative number indicating a line
                    assert addr <= 0
                    continue
                if addr not in current_iter:  # count only topmost
                    self.functions[addr] = self.functions.get(addr, 0) + 1
                    current_iter[addr] = None

    def top_profile(self):
        return [(self._get_name(k), v) for (k, v) in six.iteritems(self.functions)]

    def _get_name(self, addr):
        if self.adr_dict is not None:
            return self.adr_dict.get(addr, '<unknown code>')

        return addr

    def function_profile(self, top_function):
        """ Show functions that we call (directly or indirectly) under
        a given addr
        """
        result = {}
        total = 0
        for profile in self.profiles:
            current_iter = {}  # don't count twice
            counting = False
            for addr in profile[0]:
                if counting:
                    if addr in current_iter:
                        continue
                    current_iter[addr] = None
                    result[addr] = result.get(addr, 0) + 1
                else:
                    if addr == top_function:
                        counting = True
                        total += 1
        result = sorted(result.items(), key=lambda a: a[1])
        return result, total

    def get_top(self, profiles):
        for prof in profiles:
            if prof[0]:
                break
        else:
            raise EmptyProfileFile()
        top_addr = prof[0][0]
        top = Node(top_addr, self._get_name(top_addr))
        top.count = len(self.profiles)
        return top

    def get_tree(self):
        # fine the first non-empty profile

        top = self.get_top(self.profiles)
        addr = None
        for profile in self.profiles:
            last_addr = top.addr
            cur = top
            for i in range(0, len(profile[0])):
                if isinstance(profile[0][i], AssemblerCode):
                    continue # just ignore it for now
                addr = profile[0][i]

                if addr <= 0:
                    # negative address means line number
                    cur.lines[-addr] = cur.lines.get(-addr, 0) + 1
                else:
                    if addr == last_addr:
                        continue  # ignore duplicates
                    last_addr = addr
                    name = self._get_name(addr)
                    cur = cur.add_child(addr, name)
            if isinstance(addr, JittedCode):
                cur.meta['jit'] = cur.meta.get('jit', 0) + 1
            if isinstance(addr, NativeCode):
                cur.meta['native'] = cur.meta.get('native', 0) + 1
        # get the first "interesting" node, that is after vmprof and pypy
        # mess

        return self.filter_top(top)

    def filter_top(self, top):
        first_top = top

        while True:
            if top.name.startswith('py:<module>') and 'vmprof/__main__.py' not in top.name:
                return top
            if len(top.children) > 1:
                # pick the biggest one in case of branches in vmprof
                next = None
                count = -1
                for c in top.children.values():
                    if c.count > count:
                        count = c.count
                        next = c
                top = next
            else:
                try:
                    top = top['']
                except KeyError:
                    break

        return first_top


class Node(object):
    """ children is a dict of addr -> Node
    """
    _self_count = None
    flat = False

    def __init__(self, addr, name, count=1, children=None):
        if children is None:
            children = {}
        self.children = children
        self.name = name
        self.addr = addr
        self.count = count # starts at 1
        self.jitcodes = {}
        self.meta = {}
        self.lines = {}

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.children[item]
        for v in self.children.values():
            if item in v.name:
                return v
        raise KeyError

    def as_json(self):
        import json
        return json.dumps(self._serialize())

    def _serialize(self):
        chld = [ch._serialize() for ch in six.itervalues(self.children)]
        # if we don't make str() of addr here, JS does its
        # int -> float -> int losy convertion without
        # any warning
        return [self.name, str(self.addr), self.count, self.meta, chld]

    def _rec_count(self):
        c = 1
        for x in six.itervalues(self.children):
            c += x._rec_count()
        return c

    def walk(self, callback):
        callback(self)
        for c in six.itervalues(self.children):
            c.walk(callback)

    def cumulative_meta(self, d=None):
        if d is None:
            d = {}
        for c in six.itervalues(self.children):
            c.cumulative_meta(d)
        for k, v in six.iteritems(self.meta):
            d[k] = d.get(k, 0) + v
        return d

    def _filter(self, count):
        # XXX make a copy
        for key, c in self.children.items():
            if c.count < count:
                del self.children[key]
            else:
                c._filter(count)

    def get_self_count(self):
        if self._self_count is not None:
            return self._self_count
        self._self_count = self.count
        for elem in six.itervalues(self.children):
            self._self_count -= elem.count
        return self._self_count

    self_count = property(get_self_count)

    def add_child(self, addr, name):
        try:
            next = self.children[addr]
            next.count += 1
        except KeyError:
            next = Node(addr, name)
            self.children[addr] = next
        return next

    def __eq__(self, other):
        if not isinstance(other, Node):
            return False
        return self.name == other.name and self.addr == other.addr and self.count == other.count and self.children == other.children

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        items = sorted(self.children.items())
        child_str = ", ".join([("(%d, %s)" % (v.count, v.name))
                               for k, v in items])
        return '<Node: %s (%d) [%s]>' % (self.name, self.count, child_str)



from __future__ import print_function
from jitlog import constants as const
from jitlog import merge_point
from jitlog.objects import FlatOp, MergePoint
from vmshare.binary import (read_word, read_string,
        read_le_u16, read_le_u64,
        read_le_s64, read_bytes, read_byte,
        read_char)
import base64
import sys
import re

PARSERS = {}

def get_reader(version, mark):
    return PARSERS[mark]

def mark_parser(*versions):
    def func(parser):
        # the versions are irrelevant for now
        name = parser.__name__
        assert name.startswith("read_")
        name = "MARK_" + name[5:].upper()
        mark = getattr(const, name, -1)
        assert mark not in PARSERS, "%s has already been defined" % (parser.__name__)
        PARSERS[mark] = parser
        return parser
    return func

@mark_parser(1)
def read_resop_meta(ctx, trace, fileobj):
    assert len(ctx.forest.resops) == 0
    count = read_le_u16(fileobj)
    for i in range(count):
        opnum = read_le_u16(fileobj)
        opname = read_string(fileobj, True)
        ctx.forest.resops[opnum] = opname

@mark_parser(
     1,
    (2, "encode jitdriver name after trace_nmr")
)
def read_start_trace(ctx, trace, fileobj):
    trace_id = ctx.read_le_addr(fileobj)
    trace_type = read_string(fileobj, True)
    trace_nmr = ctx.read_le_addr(fileobj)
    # version 2 adds the jitdriver name
    if ctx.forest.version >= 2:
        jd_name = read_string(fileobj, True)
    else:
        jd_name= None
    #
    assert trace_id not in ctx.forest.traces
    ctx.forest.add_trace(trace_type, trace_id, trace_nmr, jd_name)

@mark_parser(1)
def read_trace(ctx, trace, fileobj):
    assert trace is not None
    trace_id = ctx.read_le_addr(fileobj)
    assert trace_id == trace.unique_id
    trace.start_mark(const.MARK_TRACE)

@mark_parser(1)
def read_trace_opt(ctx, trace, fileobj):
    assert trace is not None
    trace_id = ctx.read_le_addr(fileobj)
    assert trace_id == trace.unique_id
    trace.start_mark(const.MARK_TRACE_OPT)

@mark_parser(1)
def read_trace_asm(ctx, trace, fileobj):
    assert trace is not None
    trace_id = ctx.read_le_addr(fileobj)
    assert trace_id == trace.unique_id
    trace.start_mark(const.MARK_TRACE_ASM)

@mark_parser(1)
def read_input_args(ctx, trace, fileobj):
    assert trace is not None
    argnames = read_string(fileobj, True).split(',')
    trace.set_inputargs(argnames)

@mark_parser(
    1,
    (2, "encodes guard failure arguments"),
)
def read_resop(ctx, trace, fileobj):
    assert trace is not None
    opnum = read_le_u16(fileobj)
    args = read_string(fileobj, True).split(',')
    failargs = None
    if 2 <= ctx.forest.version:
        failargs = read_string(fileobj, True).split(',')
    result = args[0]
    args = args[1:]
    assert opnum in ctx.forest.resops, "opnum is not known: " + str(opnum) + \
                  " at binary pos " + str(hex(fileobj.tell()))
    opname = ctx.forest.resops[opnum]

    op = FlatOp(opnum, opname, args, result, None, -1, failargs=failargs)
    trace.add_instr(op)

TOKEN_REGEX = re.compile("TargetToken\((\d+)\)")

@mark_parser(
    1,
    (2, "encodes guard failure arguments"),
)
def read_resop_descr(ctx, trace, fileobj):
    assert trace is not None
    opnum = read_le_u16(fileobj)
    args = read_string(fileobj, True).split(',')
    descr_number = ctx.read_le_addr(fileobj)
    failargs = None
    if 2 <= ctx.forest.version:
        failargs = read_string(fileobj, True).split(',')
    descr = args[-1]
    result = args[0]
    args = args[1:-1]
    assert opnum in ctx.forest.resops, "opnum is not known: " + str(opnum) + \
                  " at binary pos " + str(hex(fileobj.tell()))
    opname = ctx.forest.resops[opnum]
    op = FlatOp(opnum, opname, args, result, descr, descr_number, failargs=failargs)
    trace.add_instr(op)

@mark_parser(1)
def read_asm_addr(ctx, trace, fileobj):
    assert trace is not None, "read asm addr, no trace obj is provided"
    addr1 = ctx.read_le_addr(fileobj)
    addr2 = ctx.read_le_addr(fileobj)
    trace.set_addr_bounds(addr1, addr2)

@mark_parser(1)
def read_asm(ctx, trace, fileobj):
    assert trace is not None, "read asm, no trace obj is provided"
    rel_pos = read_le_u16(fileobj)
    dump = read_bytes(fileobj)
    trace.set_core_dump_to_last_op(rel_pos, dump)

@mark_parser(1)
def read_init_merge_point(ctx, trace, fileobj):
    count = read_le_u16(fileobj)
    types = []
    for i in range(count):
        sem_type = read_byte(fileobj)
        gen_type = read_char(fileobj)
        d = merge_point.get_decoder(sem_type, gen_type, ctx.forest.version)
        types.append(d)
    stage = trace.get_last_stage()
    assert stage is not None, "last stage is none, but it must not be none!"
    stage.merge_point_types = types

@mark_parser(1)
def read_common_prefix(ctx, trace, fileobj):
    index = ord(fileobj.read(1))
    prefix = read_string(fileobj, True)
    stage = trace.get_last_stage()
    stage.merge_point_types[index].set_prefix(prefix)

@mark_parser(1)
def read_merge_point(ctx, trace, fileobj):
    assert trace is not None, "read merge point, trace is not None"
    stage = trace.get_last_stage()
    assert stage is not None
    #
    values = { decoder.sem_type : decoder.decode(fileobj)
               for decoder in stage.merge_point_types }
    trace.add_instr(MergePoint(values))

@mark_parser(1)
def read_stitch_bridge(ctx, trace, fileobj):
    descr_number = ctx.read_le_addr(fileobj)
    addr_tgt = ctx.read_le_addr(fileobj)
    ctx.forest.stitch_bridge(descr_number, addr_tgt)

@mark_parser(1)
def read_jitlog_counter(ctx, trace, fileobj):
    addr = ctx.read_le_addr(fileobj)
    type = read_char(fileobj)
    count = read_le_u64(fileobj)
    # entry: gets the globally numbered addr of the loop
    # bridge: gets the addr of the fail descr
    # label: gets the addr of the loop token
    trace = ctx.forest.get_trace_by_id(addr)
    if trace:
        trace.add_up_enter_count(count)
        return True
    else:
        if type == 'e':
            # it can happen that jitlog counters are present,
            # even though there is no trace to be found.
            # vm starts up (bootstrapping creates 1-2 trace),
            # jitlog is enabled afterwards
            return False
        assert type == 'b' or type == 'l'
        point_in_trace = ctx.forest.get_point_in_trace_by_descr(addr)
        if point_in_trace:
            if type == 'b':
                point_in_trace.trace.add_up_enter_count(count)
            else:
                point_in_trace.add_up_enter_count(count)
            return True
    sys.stderr.write("trace with 0x%x (type '%c' was executed %d times" \
          " but was not recorded in the log\n" % (addr, type, count))
    return False

@mark_parser(1)
def read_abort_trace(ctx, trace, fileobj):
    trace_id = ctx.read_le_addr(fileobj)
    # TODO?

@mark_parser(1)
def read_source_code(ctx, trace, fileobj):
    filename = read_string(fileobj, True)
    count = read_le_u16(fileobj)
    for i in range(count):
        lineno = read_le_u16(fileobj)
        indent = read_byte(fileobj)
        text = read_string(fileobj, True)
        ctx.forest.add_source_code_line(filename, lineno, indent, text)

@mark_parser(3)
def read_redirect_assembler(ctx, trace, fileobj):
    descr_nmr = ctx.read_le_addr(fileobj)
    new_descr_nmr = ctx.read_le_addr(fileobj)
    target = ctx.read_le_addr(fileobj)
    #
    ctx.forest.redirect_assembler(descr_nmr, new_descr_nmr, target)

@mark_parser(4)
def read_tmp_callback(ctx, trace, fileobj):
    descr_nmr = ctx.read_le_addr(fileobj)
    uniqueid = read_le_u64(fileobj) # might be -1

    ctx.forest.add_tmp_callback(descr_nmr, uniqueid)



from __future__ import absolute_import

import inspect
import linecache
import os
import six
import sys
import tokenize
import vmprof
import argparse

from vmprof.stats import EmptyProfileFile


class color(six.text_type):
    RED = '\033[31m'
    WHITE = '\033[37m'
    BLUE = '\033[94m'
    BOLD = '\033[1m'
    END = '\033[0m'

    def __new__(cls, content, color, bold=False):
        return six.text_type.__new__(
            cls, "%s%s%s%s" % (color, cls.BOLD if bold else "", content, cls.END))

class AbstractPrinter(object):
    def show(self, profile):
        """
        Read and display a vmprof profile file.

        :param profile: The filename of the vmprof profile file to display.
        :type profile: str
        """
        try:
            stats = vmprof.read_profile(profile)
        except Exception as e:
            print("Fatal: could not read vmprof profile file '{}': {}".format(profile, e))
            return

        if stats.get_runtime_in_microseconds() < 1000000:
            msg = color("WARNING: The profiling completed in less than 1 seconds. Please run your programs longer!\r\n", color.RED)
            sys.stderr.write(msg)

        try:
            tree = stats.get_tree()
            self._show(tree)
        except EmptyProfileFile as e:
            print("No stack trace has been recorded (profile is empty)!")

class PrettyPrinter(AbstractPrinter):
    """
    A pretty print for vmprof profile files.
    """

    def __init__(self, prune_percent=None, prune_level=None, indent=None):
        """

        :param prune_percent: Prune output of a profile stats node when CPU consumed
            is less than this value for the node and everything below.
        :type prune_percent: None or float
        :param prune_level: Prune output of a profile stats node when the node is deeper
            than this level down the call graph from the very top.
        :type prune_level: None or int
        :param indent: The indention per level within the call graph.
        :type indent: None or int
        """
        assert(prune_percent is None or (prune_percent >= 0 and prune_percent <= 100))
        assert(prune_level is None or (prune_level >= 0 and prune_level <= 1000))
        assert(indent is None or (indent >= 0 and indent <= 4))
        self._prune_percent = prune_percent or 0.
        self._prune_level = prune_level or 1000
        self._indent = indent or 2

    def _show(self, tree):
        self._print_tree(tree)

    def _walk_tree(self, parent, node, level, callback):
        callback(parent, node, level)
        level += 1
        if level > self._prune_level:
            return
        for c in six.itervalues(node.children):
            self._walk_tree(node, c, level, callback)

    def _print_tree(self, tree):
        total = float(tree.count)

        def print_node(parent, node, level):
            parent_name = parent.name if parent else None

            perc = round(100. * float(node.count) / total, 1)
            if parent and parent.count:
                perc_of_parent = round(100. * float(node.count) / float(parent.count), 1)
            else:
                perc_of_parent = 100.

            if perc >= self._prune_percent:
                parts = node.name.count(':')

                if parts == 3:
                    block_type, funname, funline, filename = node.name.split(':')

                    p2 = color(funname, color.BLUE, bold=True)
                    p2b = color(('.' * level * self._indent), color.BLUE)

                    p3 = []
                    if os.path.dirname(filename):
                        p3.append(color(os.path.dirname(filename) + '/', color.WHITE))
                    p3.append(color(os.path.basename(filename), color.WHITE, bold=True) + ":")
                    p3.append(color("{}".format(funline), color.WHITE))
                    p3 = ''.join(p3)

                elif parts == 1:
                    block_type, funname = node.name.split(':')
                    p2 = color("JIT code", color.RED, bold=True)
                    p2b = color('.' * level * self._indent, color.RED, bold=False)
                    p3 = color(funname, color.WHITE, bold=False)
                else:
                    p2 = color(node.name, color.WHITE)
                    p2b = color(('.' * level * self._indent), color.WHITE)
                    p3 = "<unknown>"

                p1 = color("{:>5}%".format(perc), color.WHITE, bold=True)
                p4 = color("{}%".format(perc_of_parent), color.WHITE, bold=True)

                print("{} {} {}  {}  {}".format(p1, p2b, p2, p4, p3))

        self._walk_tree(None, tree, 0, print_node)


class LinesPrinter(AbstractPrinter):
    def __init__(self, filter=None):
        self.filter = filter

    def _show(self, tree):
        for (filename, funline, funname), line_stats in self.lines_stat(tree):
            if self.filter is None or self.filter in funname or \
                                      self.filter in filename:
                self.show_func(filename, funline, funname, line_stats)

    def lines_stat(self, tree):
        funcs = {}

        def walk(node, d):
            if node is None:
                return

            parts = node.name.count(':')
            if parts == 3:
                block_type, funname, funline, filename = node.name.split(':')
                # only python supported for line profiling
                if block_type == 'py':
                    lines = d.setdefault((filename, int(funline), funname), {})
                    for l, cnt in six.iteritems(node.lines):
                        lines[l] = lines.get(l, 0) + cnt

            for c in six.itervalues(node.children):
                walk(c, d)

        walk(tree, funcs)

        return six.iteritems(funcs)

    def show_func(self, filename, start_lineno, func_name, timings, stream=None, stripzeros=False):
        """ Show results for a single function.
        """
        if stream is None:
            stream = sys.stdout

        template = '%6s %8s %8s  %-s'
        d = {}
        total_hits = 0.0

        linenos = []
        for lineno, nhits in six.iteritems(timings):
            total_hits += nhits
            linenos.append(lineno)

        if stripzeros and total_hits == 0:
            return

        stream.write("Total hits: %g s\n" % total_hits)
        if os.path.exists(filename) or filename.startswith("<ipython-input-"):
            stream.write("File: %s\n" % filename)
            stream.write("Function: %s at line %s\n" % (func_name, start_lineno))
            if os.path.exists(filename):
                # Clear the cache to ensure that we get up-to-date results.
                linecache.clearcache()
            all_lines = linecache.getlines(filename)
            try:
                sublines = inspect.getblock(all_lines[start_lineno-1:])
            except tokenize.TokenError:
                # the problem stems from getblock not being able to tokenize such an example:
                # >>> inspect.getblock(['i:i**i','}']) => TokenError
                # e.g. it fails on multi line dictionary comprehensions
                # the current approach is best effort, but cuts of some lines at the top.
                # see issue #118 for details
                sublines = all_lines[start_lineno-1:max(linenos)]
        else:
            stream.write("\n")
            stream.write("Could not find file %s\n" % filename)
            stream.write("Are you sure you are running this program from the same directory\n")
            stream.write("that you ran the profiler from?\n")
            stream.write("Continuing without the function's contents.\n")
            # Fake empty lines so we can see the timings, if not the code.
            nlines = max(linenos) - min(min(linenos), start_lineno) + 1
            sublines = [''] * nlines
        for lineno, nhits in six.iteritems(timings):
            d[lineno] = (nhits, '%5.1f' % (100* nhits / total_hits))
        linenos = range(start_lineno, start_lineno + len(sublines))
        empty = ('', '')
        header = template % ('Line #', 'Hits', '% Hits',
            'Line Contents')
        stream.write("\n")
        stream.write(header)
        stream.write("\n")
        stream.write('=' * len(header))
        stream.write("\n")
        for lineno, line in zip(linenos, sublines):
            nhits, percent = d.get(lineno, empty)
            txt = template % (lineno, nhits, percent,
                              line.rstrip('\n').rstrip('\r'))
            stream.write(txt)
            stream.write("\n")
        stream.write("\n")

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("profile")

    parser.add_argument(
        '--prune_percent',
        type=float,
        default=0,
        help="Prune output of a profile stats node below specified CPU samples.")

    parser.add_argument(
        '--prune_level',
        type=int,
        default=None,
        help='Prune output of a profile stats node below specified depth.')

    parser.add_argument(
        '--indent',
        type=int,
        default=2,
        help='The indention per level within the call graph.')

    parser.add_argument('--lines', dest='lines', action='store_true')
    parser.set_defaults(lines=False)

    parser.add_argument('--filter', dest='filter', type=str,
                        default=None, help="Filters the console output when "
                        "vmprofshow is invoked with --lines. Filters by "
                        "function names or filenames.")

    args = parser.parse_args()

    if args.lines:
        pp = LinesPrinter(filter=args.filter)
    else:
        pp = PrettyPrinter(
            prune_percent=args.prune_percent,
            prune_level=args.prune_level,
            indent=args.indent)

    pp.show(args.profile)


if __name__ == '__main__':
    main()



import os
import sys
import struct
import argparse
from collections import defaultdict
from jitlog import constants as const, merge_point

PY3 = sys.version_info[0] >= 3

class FlatOp(object):
    def __init__(self, opnum, opname, args, result,
                 descr=None, descr_number=None, failargs=None):
        self.opnum = opnum
        self.opname = opname
        self.args = args
        self.result = result
        self.descr = descr
        self.descr_number = descr_number
        self.core_dump = None
        self.failargs = failargs
        self.index = -1
        self.linkid = -1 # a unique id that is generated from the descr_number

    def get_name(self):
        return self.opname

    def getindex(self):
        return self.index

    def is_debug(self):
        return False

    def has_descr(self, descr=None):
        if not descr:
            return self.descr is not None
        return descr == self.descr_number

    def is_stitched(self):
        return False

    def get_descr_nmr(self):
        return self.descr_number

    def is_guard(self):
        return "guard" in self.opname

    def set_core_dump(self, rel_pos, core_dump):
        self.core_dump = (rel_pos, core_dump)

    def get_core_dump(self, base_addr, patches, timeval):
        coredump = self.core_dump[1][:]
        for timepos, addr, content in patches:
            if timeval < timepos:
                continue # do not apply the patch
            op_off = self.core_dump[0]
            patch_start = (addr - base_addr) - op_off 
            patch_end = patch_start + len(content)
            content_end = len(content)-1
            if patch_end >= len(coredump):
                patch_end = len(coredump)
                content_end = patch_end - patch_start
            coredump = coredump[:patch_start] + content[:content_end] + coredump[patch_end:]
        return coredump

    def __repr__(self):
        suffix = ''
        if self.result is not None:
            suffix = "%s = " % self.result
        descr = self.descr
        if descr is None:
            descr = ''
        else:
            descr = ', @' + str(descr)
        return '%s%s(%s%s)' % (suffix, self.opname,
                                ', '.join(self.args), descr)

class MergePoint(FlatOp):
    def __init__(self, values):
        assert isinstance(values, dict)
        self.values = values

    def get_name(self):
        return ""

    def is_debug(self):
        return True

    def get_scope(self):
        scope = const.MP_SCOPE[0]
        if scope in self.values:
            return self.values[scope]
        return ""

    def get_source_line(self):
        filename = None
        lineno = None
        for sem_type, value in self.values.items():
            if sem_type == const.MP_FILENAME[0]:
                filename = value
            if sem_type == const.MP_LINENO[0]:
                lineno = value
        if filename is None or lineno is None:
            return 0, None
        return lineno, filename

    def has_descr(self, descr=None):
        return False

    def set_core_dump(self, rel_pos, core_dump):
        raise NotImplementedError

    def get_core_dump(self, base_addr, patches, timeval):
        raise NotImplementedError

    def __repr__(self):
        return 'debug_merge_point(xxx)'

class Stage(object):
    def __init__(self, mark, timeval):
        self.mark = mark
        self.ops = []
        self.merge_points = []
        self.timeval = timeval
        self.stitch_points = []

    def getoperations(self):
        return self.ops

    def get_last_op(self):
        if len(self.ops) == 0:
            return None
        return self.ops[-1]

    def get_op(self, i):
        if i < 0 or len(self.ops) <= i:
            return None
        return self.ops[i]

    def append_op(self, op):
        op.index = len(self.ops)
        if isinstance(op, MergePoint):
            self.merge_points.append(op)
        else:
            self.ops.append(op)

    def get_ops(self):
        return self.ops

    def get_merge_points(self):
        return self.merge_points

    def get_opcount(self):
        return len(self.ops)

    def is_asm(self):
        return self.mark == 'asm'

    def __repr__(self):
        return 'Stage("%s", %d ops)' % (self.mark, self.get_opcount())

from collections import namedtuple

class TraceLink(namedtuple('TraceLink', 'origin target')):
    pass

class Trace(object):
    def __init__(self, forest, trace_type, tick, unique_id, jd_name=None):
        self.forest = forest
        self.jd_name = jd_name
        self.type = trace_type
        self.inputargs = []
        assert self.type in ('loop', 'bridge')
        self.unique_id = unique_id
        self.stages = {}
        self.last_mark = None
        self.addrs = (-1,-1)
        # this saves a quadrupel for each
        self.my_patches = None
        self.counter = 0
        self.point_counters = {}
        self.merge_point_files = defaultdict(list)
        self.descr_number = 0 # the descr this trace is attached to
        # the trace links that lead to this trace
        self.links_up = []
        # the links leading from this trace
        self.links = []

    def link(self, op, target):
        self.links.append(TraceLink(PointInTrace(self, op), PointInTrace(target, None)))

    def backward_link(self, orig):
        """ connect parent with this trace. orig is the point
        in the parent trace, enter the first op in this trace """
        assert isinstance(orig, PointInTrace)
        enter = PointInTrace(self, None)
        self.links_up.append(TraceLink(orig, enter))

    def get_id(self):
        """ Return the unique id for that trace object """
        return self.unique_id

    def add_up_enter_count(self, count):
        self.counter += count

    def get_counter_points(self):
        d = {0: self.counter}
        d.update(self.point_counters)
        return d

    def get_stitched_descr_number(self):
        return self.descr_number

    def get_parent(self):
        for link in self.links_up:
            if link.origin and link.target.op is None:
                return link.origin.trace
        return None

    def get_first_merge_point(self):
        stage = self.get_stage('opt')
        if stage:
            mps = stage.get_merge_points()
            if len(mps) != 0:
                return mps[0]
        return None

    def get_stage(self, type):
        assert type is not None
        return self.stages.get(type, None)

    def start_mark(self, mark):
        mark_name = 'noopt'
        if mark == const.MARK_TRACE_OPT:
            mark_name = 'opt'
        elif mark == const.MARK_TRACE_ASM:
            mark_name = 'asm'
        else:
            assert mark == const.MARK_TRACE
            if self.last_mark == mark_name:
                # NOTE unrolling
                #
                # this case means that the optimizer has been invoked
                # twice (see compile_loop in rpython/jit/metainterp/compile.py)
                # and the loop was unrolled in between.
                #
                # we just return here, which means the following ops will just append the loop
                # ops to the preamble ops to the current stage!
                return
        self.last_mark = mark_name
        assert mark_name is not None
        if mark_name in self.stages:
            return self.stages[mark_name]
        tick = self.forest.timepos
        stage = Stage(mark_name, tick)
        self.stages[mark_name] = stage
        return stage

    def get_last_stage(self):
        return self.stages.get(self.last_mark, None)

    def set_core_dump_to_last_op(self, rel_pos, dump):
        assert self.last_mark is not None
        flatop = self.get_stage(self.last_mark).get_last_op()
        flatop.set_core_dump(rel_pos, dump)

    def add_instr(self, op):
        stage = self.get_stage(self.last_mark)
        stage.append_op(op)
        if stage.is_asm():
            if op.has_descr():
                stage.stitch_points.append(PointInTrace(self, op))
                nmr = op.get_descr_nmr()
                if nmr == 0x0:
                    sys.stderr.write("descr in trace %s should not be 0x0\n" % self)
                else:
                    dict = self.forest.descr_nmr_to_point_in_trace
                    # a label could already reside in that position
                    if nmr not in dict:
                        dict[nmr] = PointInTrace(self, op)
                    else:
                        pass
                        #sys.stderr.write("duplicate descr: 0x%x\n" % nmr)

                if op.get_name() == "label":
                    self.forest.labels[nmr] = PointInTrace(self, op)
                if op.get_name() == "jump":
                    self.forest.jumps[nmr] = PointInTrace(self, op)

        if op.get_name() == "increment_debug_counter":
            prev_op = stage.get_op(op.index-1)
            # look for the previous operation, it is a label saved
            # in descr_number_to_point_in_trace
            if prev_op:
                descr_number = prev_op.get_descr_nmr()
                pit = self.forest.get_point_in_trace_by_descr(descr_number)
                pit.set_inc_op(op)

        if isinstance(op, MergePoint):
            lineno, filename = op.get_source_line()
            if filename:
                self.merge_point_files[filename].append(lineno)

    def is_bridge(self):
        return self.type == 'bridge'

    def set_inputargs(self, args):
        self.inputargs = args

    def set_addr_bounds(self, a, b):
        self.addrs = (a,b)
        if a in self.forest.addrs:
            sys.stderr.write("jit log sets address bounds to a location another trace already is resident of\n")
        self.forest.addrs[a] = self

    def is_assembled(self):
        """ return True if the jit log indicated to have assembled this trace """
        return self.addrs[0] != -1

    def get_addrs(self):
        return tuple(self.addrs)

    def contains_addr(self, addr):
        return self.addrs[0] <= addr <= self.addrs[1]

    def contains_patch(self, addr):
        if self.addrs is None:
            return False
        return self.addrs[0] <= addr <= self.addrs[1]

    def get_core_dump(self, timeval=-1, opslice=(0,-1)):
        if timeval == -1:
            timeval = 2**31-1 # a very high number
        if self.my_patches is None:
            self.my_patches = []
            for patch in self.forest.patches:
                patch_time, addr, content = patch
                if self.contains_patch(addr):
                    self.my_patches.append(patch)

        core_dump = []
        start,end = opslice
        if end == -1:
            end = len(opslice)
        ops = None
        stage = self.get_stage('asm')
        if not stage:
            return None # no core dump!
        for i, op in enumerate(stage.get_ops()):
            if start <= i <= end:
                dump = op.get_core_dump(self.addrs[0], self.my_patches, timeval)
                core_dump.append(dump)
        return ''.join(core_dump)

    def get_name(self):
        stage = self.get_stage('opt')
        if not stage:
            pass
        return 'unknown'

    def get_failing_guard(self):
        """ return the resoperation (FlatOp) for the descr this trace is attached to """
        if len(self.links_up) > 0:
            for link in self.links_up:
                origin = link.origin
                if origin and origin.op:
                    return origin.op
        return None

    def __repr__(self):
        return 'Trace(%d, 0x%x, %d)' % (self.unique_id, id(self), len(self.stages))

def iter_ranges(numbers):
    if len(numbers) == 0:
        return
    numbers.sort()
    first = numbers[0]
    last = numbers[0]
    for pos, i in enumerate(numbers[1:]):
        if (i - first) > 50:
            yield range(first, last+1)
            if pos+1 < len(numbers):
                last = i
                first = i
            else:
                return
        else:
            last = i
    yield range(first, last+1)

class PointInTrace(object):
    def __init__(self, trace, op):
        self.trace = trace
        self.op = op
        self.inc_op = None

    def enters_the_trace(self):
        return self.op is None

    def get_operation(self):
        return self.op

    def set_inc_op(self, op):
        if self.inc_op is None:
            self.inc_op = op

    def add_up_enter_count(self, count):
        if not self.inc_op:
            return False# this is a label!
        counters = self.trace.point_counters
        i = self.inc_op.index
        c = counters.get(i, 0)
        counters[i] = c + count
        return True

    def __repr__(self):
        return "PointInTrace(%s, op %s)" % (self.trace, self.op)

def decode_source(source_bytes):
    # copied from _bootstrap_external.py
    """Decode bytes representing source code and return the string.
    Universal newline support is used in the decoding.
    """
    import _io
    import tokenize  # To avoid bootstrap issues.
    source_bytes_readline = _io.BytesIO(source_bytes).readline
    encoding = tokenize.detect_encoding(source_bytes_readline)
    newline_decoder = _io.IncrementalNewlineDecoder(None, True)
    return newline_decoder.decode(source_bytes.decode(encoding[0]))

def read_python_source(file):
    with open(file, 'rb') as fd:
        data = fd.read()
        if PY3:
            data = decode_source(data)
        return data

class TraceForest(object):
    def __init__(self, version, is_32bit=False, machine=None):
        self.word_size = 4 if is_32bit else 8
        self.version = version
        self.machine = machine
        self.roots = []
        self.traces = {}
        self.addrs = {}
        self.last_trace = None
        self.resops = {}
        self.timepos = 0
        self.patches = []
        self.stitches = {}
        self.filepath = None
        # a mapping from source file name -> {lineno: (indent, line)}
        self.source_lines = defaultdict(dict)
        self.descr_nmr_to_point_in_trace = {}
        self.exc = None # holds an exception object if an error occured
        self.labels = {}
        self.jumps = {}
        self.redirect_descrs = {}
        self.linkid = 0

    def next_link_id(self):
        linkid = self.linkid
        self.linkid += 1
        return linkid

    def exception_raised(self):
        return self.exc

    def unlink_jitlog(self):
        if self.filepath and os.path.exists(self.filepath):
            os.unlink(self.filepath)
            self.filepath = None

    def get_point_in_trace_by_descr(self, descr):
        assert isinstance(descr, int)
        return self.descr_nmr_to_point_in_trace.get(descr, None)

    def get_source_line(self, filename, lineno):
        lines = self.source_lines.get(filename, None)
        if not lines:
            return None, None
        return lines.get(lineno, (None, None))

    def copy_and_add_source_code_tags(self):
        with open(self.filepath, "ab") as fd:
            blob = self.encode_source_code_lines()
            fd.write(blob)

    def extract_source_code_lines(self):
        if len(self.source_lines) > 0:
            # uploading the jitlog more than once, will append source code
            # to the jitlog, thus it will yield many duplicates. If there
            # is already source code attached, we skip this step
            return False
        file_contents = {}
        for _, trace in self.traces.items():
            for file, lines in trace.merge_point_files.items():
                if file not in file_contents:
                    if not os.path.exists(file):
                        continue
                    code = read_python_source(file)
                    file_contents[file] = code.splitlines()

                split_lines = file_contents[file]
                saved_lines = self.source_lines[file]
                for int_range in iter_ranges(lines):
                    for r in int_range:
                        line = split_lines[r-1]
                        data = line.lstrip()
                        diff = len(line) - len(data)
                        indent = diff
                        for i in range(0, diff):
                            if line[i] == '\t':
                                indent += 7
                        saved_lines[r] = (indent, data)
        return True

    def get_trace(self, id):
        return self.traces.get(id, None)

    def get_trace_by_addr(self, addr):
        return self.addrs.get(addr, None)

    def get_trace_by_id(self, id):
        return self.traces.get(id, None)

    def add_trace(self, trace_type, unique_id, trace_nmr, jd_name=None):
        """ Create a new trace object and attach it to the forest """
        trace = Trace(self, trace_type, self.timepos, unique_id, jd_name)
        trace.stamp = len(self.traces)
        self.traces[unique_id] = trace
        self.last_trace = trace
        return trace

    def stitch_bridge(self, descr_number, addr_to):
        assert isinstance(descr_number, int)
        bridge = self.get_trace_by_addr(addr_to)
        if bridge is None:
            raise Exception("bridge is None")
        #assert bridge.descr_number == 0, "a bridge can only be stitched once"
        #bridge.descr_number = descr_number
        # TODO remove this
        self.stitches[descr_number] = bridge.unique_id
        assert bridge is not None, ("no trace to be found for addr 0x%x" % addr_to)
        point_in_trace = self.get_point_in_trace_by_descr(descr_number)
        if not point_in_trace:
            sys.stderr.write("link to trace of descr 0x%x not found!\n" % descr_number)
        else:
            trace = point_in_trace.trace
            op = point_in_trace.op
            trace.link(op, bridge)
            bridge.backward_link(point_in_trace)

    def get_stitch_target(self, descr_number):
        assert isinstance(descr_number, int)
        return self.stitches.get(descr_number)

    def patch_memory(self, addr, content, timeval):
        self.patches.append((timeval, addr, content))

    def time_tick(self):
        self.timepos += 1

    def is_jitlog_marker(self, marker):
        if len(marker) == 0:
            return False
        assert len(marker) == 1
        return const.MARK_JITLOG_START <= marker <= const.MARK_JITLOG_END

    def encode_source_code_lines(self):
        marks = []
        for filename, lines in self.source_lines.items():
            marks.append(const.MARK_SOURCE_CODE)
            data = filename
            if PY3:
                data = data.encode('utf-8')
            marks.append(struct.pack('<I', len(data)))
            marks.append(data)

            marks.append(struct.pack('<H', len(lines)))
            for lineno, (indent, line) in lines.items():
                marks.append(struct.pack('<HBI', lineno, indent, len(line)))
                marks.append(line.encode('utf-8'))
        return b''.join(marks)

    def add_source_code_line(self, filename, lineno, indent, line):
        dict = self.source_lines[filename]
        if lineno in dict:
            sys.stderr.write("dup source code. %s line %d\n" % (filename, lineno))
        dict[lineno] = (indent, line)

    def add_tmp_callback(self, descr_nmr, uid=-1):
        self.redirect_descrs[descr_nmr] = uid

    def redirect_assembler(self, descr_number, new_descr_number, trace_id):
        assert isinstance(descr_number, int)
        self.stitches[descr_number] = trace_id
        if descr_number not in self.redirect_descrs:
            pass
        else:
            del self.redirect_descrs[descr_number]
        # setup the object properties
        trace = self.get_trace_by_id(trace_id)
        if trace is None:
            sys.stderr.write("could not redirect_assembler 0x%x 0x%x id 0x%x\n" % (descr_number, new_descr_number, trace_id))
            return
        point_in_trace = self.get_point_in_trace_by_descr(descr_number)
        if not point_in_trace:
            sys.stderr.write("redirect call assembler: link to trace of descr 0x%x not found trace %d!\n" % (descr_number, trace_id))
        else:
            parent = point_in_trace.trace
            op = point_in_trace.op
            parent.link(op, trace)
            trace.backward_link(point_in_trace)



# generated constants from rpython/rlib/jitlog.py
import struct
MARK_JITLOG_START = struct.pack("b", 0x10)
MARK_INPUT_ARGS = struct.pack("b", 0x11)
MARK_RESOP_META = struct.pack("b", 0x12)
MARK_RESOP = struct.pack("b", 0x13)
MARK_RESOP_DESCR = struct.pack("b", 0x14)
MARK_ASM_ADDR = struct.pack("b", 0x15)
MARK_ASM = struct.pack("b", 0x16)
MARK_TRACE = struct.pack("b", 0x17)
MARK_TRACE_OPT = struct.pack("b", 0x18)
MARK_TRACE_ASM = struct.pack("b", 0x19)
MARK_STITCH_BRIDGE = struct.pack("b", 0x1a)
MARK_START_TRACE = struct.pack("b", 0x1b)
MARK_JITLOG_COUNTER = struct.pack("b", 0x1c)
MARK_INIT_MERGE_POINT = struct.pack("b", 0x1d)
MARK_JITLOG_HEADER = struct.pack("b", 0x1e)
MARK_MERGE_POINT = struct.pack("b", 0x1f)
MARK_COMMON_PREFIX = struct.pack("b", 0x20)
MARK_ABORT_TRACE = struct.pack("b", 0x21)
MARK_SOURCE_CODE = struct.pack("b", 0x22)
MARK_REDIRECT_ASSEMBLER = struct.pack("b", 0x23)
MARK_TMP_CALLBACK = struct.pack("b", 0x24)
MARK_JITLOG_END = struct.pack("b", 0x25)
MP_INDEX = (0x4,"i")
MP_SCOPE = (0x8,"s")
MP_FILENAME = (0x1,"s")
MP_OPCODE = (0x10,"s")
MP_LINENO = (0x2,"i")
MP_STR = (0x0,"s")
MP_INT = (0x0,"i")
SEM_TYPE_NAMES = {
    0x4: "index",
    0x8: "scope",
    0x1: "filename",
    0x10: "opcode",
    0x2: "lineno",
}



import sys
import struct
import traceback
from jitlog import constants as const, marks
from jitlog.objects import TraceForest
from vmshare.binary import read_string
from io import BytesIO
from collections import defaultdict

JITLOG_MIN_VERSION = 1
JITLOG_VERSION = 1

class ParseContext(object):
    def __init__(self, forest):
        self.descrs = defaultdict(list)
        self.forest = forest
        self.word_size = self.forest.word_size

    def read_le_addr(self, fileobj):
        b = fileobj.read(self.word_size)
        if self.word_size == 4:
            res = int(struct.unpack('I', b)[0])
        else:
            res = int(struct.unpack('Q', b)[0])
        return res


class ParseException(Exception):
    pass

def read_jitlog_data(filename):
    with open(str(filename), 'rb') as fileobj:
        return fileobj.read()

def parse_jitlog(filename, data=None):
    if data is None:
        data = read_jitlog_data(filename)
    fileobj = BytesIO(data)
    f = _parse_jitlog(fileobj)
    f.filepath = filename
    return f

def _parse_jitlog(fileobj):
    """ JitLog is parsed. if an exception is raised after
    the header has been read, it is saved in the field 'exc' on
    the forest returned.

    This allows to parse an incomplete log file an still display
    everything that was readable!
    If a ParseException is directly raised, this is an error that cannot
    be recovered from!
    """
    try:
        is_jit_log = fileobj.read(1) == const.MARK_JITLOG_HEADER
        version = ord(fileobj.read(1)) | (ord(fileobj.read(1)) << 8)
        is_32bit = ord(fileobj.read(1))
        machine = read_string(fileobj, True)
        forest = TraceForest(version, is_32bit, machine)
        ctx = ParseContext(forest)
    except Exception:
        raise ParseException("Header malformed")
    #
    if not is_jit_log:
        raise ParseException("Missing header. Provided input might not be a jitlog!")
    if version < JITLOG_MIN_VERSION:
        raise ParseException("Version %d is not supported" % version)
    while True:
        marker = fileobj.read(1)
        if len(marker) == 0:
            break # end of file!
        if not forest.is_jitlog_marker(marker):
            msg = "marker unknown: 0x%x at pos 0x%x" % \
                      (ord(marker), fileobj.tell())
            forest.exc = ParseException(msg)
            break
        trace = forest.last_trace
        try:
            read = marks.get_reader(version, marker)
            read(ctx, trace, fileobj)
            forest.time_tick()
        except KeyError as e:
            forest.exc = e
            break
        except ParseException as e:
            forest.exc = e
            break
        except Exception as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            tb = traceback.extract_tb(exc_traceback, limit=3)
            msg = "failed at 0x%x with marker %s with exc \"%s\". trace back: \"%s\"" %\
                    (fileobj.tell(), marker, str(e), tb)
            forest.exc = ParseException(msg)
            break

    return forest




from colorama import init, deinit, Fore, Back, Style

class PrettyPrinter(object):
    def __enter__(self):
        pass

    def __exit__(self):
        pass

    def op(self, op):
        suffix = ''
        if op.result is not None and op.result != '?':
            suffix = "%s = " % self.var(op.result)
        descr = op.descr
        if descr is None:
            descr = ''
        else:
            descr = ', @' + self.descr(descr)
        args = [self.var(arg) for arg in op.args]
        return '%s%s(%s%s)' % (suffix, self.opname(op.opname),
                                ', '.join(args), descr)

    def trace(self, fd, trace):
        for name, stage in trace.stages.items():
            fd.write(self.stage_name(stage))
            fd.write("\n")
            for op in stage.getoperations():
                fd.write('  ' + self.op(op) + '\n')

    def var(self, var):
        return var

    def descr(self, descr):
        return descr

    def opname(self, opname):
        return opname

    def stage_name(self, stage):
        return str(stage)

class ColoredPrettyPrinter(PrettyPrinter):
    def __enter__(self):
        init()
        return self

    def __exit__(self, a, b, c):
        deinit()

    def opname(self, opname):
        return Fore.CYAN + opname + Style.RESET_ALL

    def var(self, var):
        if len(var) > 0:
            if var[0] == 'i':
                return Fore.BLUE + var + Style.RESET_ALL
            if var[0] == 'p' or var[0] == 'r':
                return Fore.GREEN + var + Style.RESET_ALL
            if var[0] == 'f':
                return Fore.BLUE + var + Style.RESET_ALL
        return Fore.RED + var + Style.RESET_ALL




import platform
import argparse
from vmshare.service import Service

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("profile")
    parser.add_argument("--web-url", default='http://vmprof.com', help='Target host')
    parser.add_argument("--web-auth", default=None, help='Authtoken for your acount on the server')
    args = parser.parse_args()

    host, auth = args.web_url, args.web_auth
    filename = args.profile
    service = Service(host, auth)
    interpname = platform.python_implementation()
    service.post({ Service.FILE_CPU_PROFILE: filename,
                   Service.FILE_JIT_PROFILE: filename + '.jit',
                   'argv': interpname + ' -m vmprof.upload ' + filename,
                   'VM': interpname })

if __name__ == '__main__':
    main()



import sys
import json
import argparse
import os
import requests
import vmprof
import tempfile
from jitlog.parser import read_jitlog_data, parse_jitlog
from vmprof.stats import Stats
from vmprof.stats import EmptyProfileFile
import jitlog
import gzip
import warnings

PY3 = sys.version_info[0] >= 3

class ServiceException(Exception):
    pass

def is_errornous(json):
    return 'detail' in json

def service_exception_from_json(json):
    assert 'error' in json
    return ServiceException(json['error'])


def compress_file(filename):
    fileno, name = tempfile.mkstemp(suffix='.zip')
    os.close(fileno)
    with open(filename, 'rb') as fd:
        with gzip.open(name, 'wb') as zipfd:
            while True:
                chunk = fd.read(1024)
                if not chunk:
                    break
                zipfd.write(chunk)
    return name

class Service(object):
    FILE_CPU_PROFILE = 'cpu'
    FILE_MEM_PROFILE = 'mem'
    FILE_JIT_PROFILE = 'jit'

    def __init__(self, host, auth=None):
        self.host = host
        self.auth = auth
        self.runtime_id = None
        self.csrf_token = None

    def get_headers(self):
        base_headers = { 'Content-Type': 'application/json' }
        if self.auth:
            base_headers['AUTHORIZATION'] = "Token %s" % self.auth
        if self.csrf_token:
            base_headers['csrftoken'] = self.csrf_token
        return base_headers

    def get_url(self, path):
        host = self.host
        path = path.lstrip('/')
        if host.startswith("http"):
            url = '%s/%s' % (host.rstrip("/"), path)
        else:
            url = 'http://%s/%s' % (host.rstrip("/"), path)
        return url

    def stop_if_error_occured(self, response):
        if response.status_code != 200:
            sys.stderr.write("server rejected meta data." \
                             " status: %d, msg: '%s'\n" % \
                             (response.status_code, response.text))
            raise ServiceException()

    def post_new_entry(self, data={}):
        url = self.get_url('/api/runtime/new')
        headers = self.get_headers()
        bytesdata = json.dumps(data).encode('utf-8')
        response = requests.post(url, data=bytesdata, headers=headers)
        self.stop_if_error_occured(response)
        j = response.json()
        return j.get('runtime_id', None)

    def post_file(self, rid, filename, filetype, compress=False):
        if not os.path.exists(filename):
            return None

        if compress:
            filename = compress_file(filename)

        if 'vmprof.com' in self.host:
            # not sure about nginx, what does MB mean? 1MB == 1000 byte or 1024 byte?
            # take the lower bound...
            size = os.path.getsize(filename)
            if size >= 100*1000*1000: # 100 MB
                warnings.warn("WARNING: vmprof.com limits the "
                              "compressed profile file to 100 MBs. "
                              "The upload might fail!\n")
        with open(filename, 'rb') as fd:
            url = self.get_url('/api/runtime/upload/%s/%s/add' % (filetype, rid))
            headers = self.get_headers()
            headers['Content-Disposition'] = 'attachment; filename='+filename
            del headers['Content-Type']
            response = requests.post(url, headers=headers, data=fd)
            self.stop_if_error_occured(response)
            return response

    def post(self, kwargs):
        sys.stderr.write("Uploading to %s...\n" % self.host)

        argv = kwargs.get('argv', '-')
        vm = kwargs.get('VM','unknown')

        rid = self.post_new_entry({'argv': argv, 'VM': vm})
        if rid is None:
            raise ServiceException("could not create meta data for profiles")

        if Service.FILE_CPU_PROFILE in kwargs:
            filename = kwargs[Service.FILE_CPU_PROFILE]
            if os.path.exists(filename):
                sys.stderr.write(" => Uploading the cpu profile...\n")
                self.post_file(rid, filename,
                               Service.FILE_CPU_PROFILE, compress=False)
                sys.stderr.write('      ' + self.get_url('#/%s' % rid) + "\n")
        if Service.FILE_JIT_PROFILE in kwargs:
            filename = kwargs[Service.FILE_JIT_PROFILE]
            if os.path.exists(filename):
                sys.stderr.write(" => Uploading the jit log...\n")
                forest = parse_jitlog(filename)
                if forest.exception_raised():
                    sys.stderr.write(" error: %s\n" % forest.exception_raised())
                # append source code to the binary
                forest.extract_source_code_lines()
                forest.copy_and_add_source_code_tags()
                filename = forest.filepath
                response = self.post_file(rid, filename,
                               Service.FILE_JIT_PROFILE, compress=True)
                forest.unlink_jitlog()
                json = response.json()
                if 'jid' in json:
                    url = self.get_url('#/%s/traces' % json['jid'])
                else:
                    url = self.get_url('#/%s' % rid)
                sys.stderr.write('      ' + url + "\n")

        self.finalize_entry(rid)


    def finalize_entry(self, rid, data=b""):
        url = self.get_url('/api/runtime/%s/freeze/' % rid)
        headers = self.get_headers()
        response = requests.post(url, data=data, headers=headers)
        self.crsf_token = None
        if response.status_code != 200:
            sys.stderr.write("server failed to freeze these runtime profiles." \
                             " status: %d, msg: '%s'\n" % \
                                (response.status_code, response.text))
            return False
        return True

